#include "stdafx.h"

#include "AnimData.h"
#include "AnimKey.h"
#include "AnimTrack.h"


CAnimTrack::CAnimTrack(DWORD dwNumKeys, DWORD dwNumFrames, DWORD dwFrameRate, FLOAT fAnimFPS) : 
    m_dwNumKeys(dwNumKeys), m_dwNumFrames(dwNumFrames), m_dwFrameRate(dwFrameRate), m_fAnimFPS(fAnimFPS),
    m_fOverallSpeed((FLOAT)1.0), m_dwNextAnimFrame(0), m_boActivateInterpolation(true), m_ListKeys(),
    m_dwCurKey(0), m_fCurFrame((FLOAT)0.0), m_fInterpFactor((FLOAT)0.0), m_boNextKeyIsInCurrentAnim(true)
{
    ASSERT(dwNumKeys > 0);
    ASSERT(dwNumFrames >= dwNumKeys);
    ASSERT(dwFrameRate > 0);
    ASSERT(fAnimFPS >= (FLOAT)0.0);

    m_lpNextAnim        = NULL;

    m_dwRemainingFrames = m_dwNumFrames - ((m_dwNumFrames - 1) / m_dwFrameRate) * m_dwFrameRate;
}


void    CAnimTrack::vSetNextAnim(CAnimTrack* lpAnim, DWORD dwFrame)
{
    ASSERT(lpAnim);
    //ASSERT(dwFrame < lpAnim->dwGetNumFrames()  &&  dwFrame >= 0);

    if(dwFrame >= lpAnim->dwGetNumFrames()) dwFrame = 0;

    m_lpNextAnim        = lpAnim;
    m_dwNextAnimFrame   = dwFrame;
}


bool    CAnimTrack::boSetCurrentFrame(FLOAT fFrame)
{
    m_fCurFrame = fFrame;

    // Have we crossed over the end of the animation ?
    if((DWORD)m_fCurFrame >= m_dwNumFrames)
        {
        // Yes. Is the next animation not the current one ?
        if(m_lpNextAnim != this)
            {
            // Yes. The caller must handle this situation
            FLOAT   fCurKey = fGetKeyFromFrame(m_fCurFrame);

            m_fInterpFactor = m_boActivateInterpolation ? fCurKey - (DWORD) fCurKey : (FLOAT) 0.0;
            return false;
            }
        else
            {
            // No. Reset internal variables
            m_fCurFrame -= m_dwNumFrames;

            m_fCurFrame += m_dwNextAnimFrame;
            
            // The next test should always fail, but if the frame rate is *very* bad
            // (fElapsedTime is huge), it could succeed
            if(m_fCurFrame >= m_dwNumFrames) m_fCurFrame = (FLOAT) (m_dwNumFrames - 1);
            }
        }

    // Calculate the current key
    FLOAT   fCurKey = fGetKeyFromFrame(m_fCurFrame);

    m_dwCurKey = (DWORD) fCurKey;

    // Use the next key for the interpolation. Handle the case where the
    // next key is in the next animation and not in the current one
    DWORD   dwNextFrame = (DWORD) ((m_dwCurKey + 1) * m_dwFrameRate);
    FLOAT   fSpeedFactor = (FLOAT) 1.0;

    if(dwNextFrame < m_dwNumFrames)
        {
        // The key to use for the interpolation is the next one in the current animation
        m_boNextKeyIsInCurrentAnim = true;
        }
    else
        {
        // The next key is in the next animation
        m_boNextKeyIsInCurrentAnim = false;

        // m_fSpeedFactor is used to correctly interpolate the last key of the animation with
        // the first one of the next animation. Indeed, interpolation would failed if the number
        // of frames after the last key of the current animation is not equal to the animation
        // m_dwFrameRate (which is possible)
        fSpeedFactor = ((FLOAT)m_dwFrameRate) / m_dwRemainingFrames;
        }

    if(m_boActivateInterpolation)
        {
        m_fInterpFactor = (fCurKey - m_dwCurKey) * fSpeedFactor;
        }
    else
        {
        m_fInterpFactor = (FLOAT) 0.0;
        }

    return true;
}


bool    CAnimTrack::boRunForward(FLOAT fElapsedTime)
{
    m_fCurFrame += (fElapsedTime * m_fAnimFPS) * m_fOverallSpeed;

    return boSetCurrentFrame(m_fCurFrame);
}


void    CAnimTrack::vGetInterpolatedData(DWORD dwNumData,CAnimData* lpData,bool boDetectRecursInfinite)
{
    ASSERT(m_dwCurKey < m_dwNumKeys);
    ASSERT(m_fInterpFactor <= 1.0f);
    ASSERT(lpData != NULL);

    const CAnimData&    DataCurKey  = m_ListKeys[m_dwCurKey].GetData(dwNumData);
    CAnimData           DataInterpolated;

    if(m_boNextKeyIsInCurrentAnim)
        {
        DWORD dwNextKey = m_dwCurKey + 1;

        if(dwNextKey >= m_dwNumKeys) dwNextKey = 0; // to handle animations with only one key

        CAnimData::vInterpolate(DataCurKey,m_ListKeys[dwNextKey].GetData(dwNumData),lpData,m_fInterpFactor);
        }
    else
        {
        if(m_dwNumKeys == 1)
            {
            CAnimData::vInterpolate(DataCurKey,DataCurKey,lpData,m_fInterpFactor);
            }
        else
            {
            if(boDetectRecursInfinite)
                {
                CAnimData::vInterpolate(DataCurKey,m_ListKeys[m_dwNumKeys - 1].GetData(dwNumData),lpData,m_fInterpFactor);
                }
            else
                {
                ASSERT(m_lpNextAnim != NULL);

                m_lpNextAnim->vPushState();
                m_lpNextAnim->boSetCurrentFrame((FLOAT)m_dwNextAnimFrame);
                m_lpNextAnim->vGetInterpolatedData(dwNumData,&DataInterpolated,true);
                m_lpNextAnim->vPopState();

                CAnimData::vInterpolate(DataCurKey,DataInterpolated,lpData,m_fInterpFactor);
                }
            }
        }
}


void    CAnimTrack::vGetInterpolatedBB(CBoundingBox* lpBB,bool boDetectRecursInfinite)
{
    ASSERT(m_dwCurKey < m_dwNumKeys);
    ASSERT(m_fInterpFactor <= 1.0f);
    ASSERT(lpBB != NULL);

    const CBoundingBox& BBCurKey  = m_ListKeys[m_dwCurKey].GetBB();
    CBoundingBox        BBInterpolated;

    if(m_boNextKeyIsInCurrentAnim)
        {
        DWORD dwNextKey = m_dwCurKey + 1;

        if(dwNextKey >= m_dwNumKeys) dwNextKey = 0; // to handle animations with only one key

        CBoundingBox::vInterpolate(BBCurKey,m_ListKeys[dwNextKey].GetBB(),lpBB,m_fInterpFactor);
        }
    else
        {
        if(m_dwNumKeys == 1)
            {
            CBoundingBox::vInterpolate(BBCurKey,BBCurKey,lpBB,m_fInterpFactor);
            }
        else
            {
            if(boDetectRecursInfinite)
                {
                CBoundingBox::vInterpolate(BBCurKey,m_ListKeys[m_dwNumKeys - 1].GetBB(),lpBB,m_fInterpFactor);
                }
            else
                {
                ASSERT(m_lpNextAnim != NULL);

                m_lpNextAnim->vPushState();
                m_lpNextAnim->boSetCurrentFrame((FLOAT)m_dwNextAnimFrame);
                m_lpNextAnim->vGetInterpolatedBB(&BBInterpolated,true);
                m_lpNextAnim->vPopState();

                CBoundingBox::vInterpolate(BBCurKey,BBInterpolated,lpBB,m_fInterpFactor);
                }
            }
        }
}
