// MarkupSTL.cpp: implementation of the CMarkupSTL class.
//
// Markup Release 6.1
// Copyright (C) 1999-2001 First Objective Software, Inc. All rights reserved
// Go to www.firstobject.com for the latest CMarkup and EDOM documentation
// Use in commercial applications requires written permission
// This software is provided "as is", with no warranty.
#include "stdafx.h"

#include "MarkupSTL.h"

using namespace std;

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

// To ignore case, define MARKUP_IGNORECASE
#define MARKUP_IGNORECASE

#ifdef MARKUP_IGNORECASE
#define x_STRMATCH(s1,s2) (stricmp(s1.c_str(),s2)==0)
#else
#define x_STRMATCH(s1,s2) (strcmp(s1.c_str(),s2)==0)
#endif


void CMarkupSTL::operator=( const CMarkupSTL& markup )
{
	m_iPos = markup.m_iPos;
	m_iPosChild = markup.m_iPosChild;
	m_iPosFree = markup.m_iPosFree;
	m_aPos = markup.m_aPos;
	m_nLevel = markup.m_nLevel;
	m_strDoc = markup.m_strDoc;
}

void CMarkupSTL::ResetPos()
{
	// Reset the main and child positions
	m_iPos = 0;
	m_iPosChild = 0;
	m_nLevel = 0;
};

bool CMarkupSTL::SetDoc( const char* szDoc )
{
	// Reset indexes
	m_iPosFree = 1;
	ResetPos();
	m_mapSavedPos.clear();

	// Set document text
	if ( szDoc )
		m_strDoc = szDoc;
	else
		m_strDoc.erase();

	// Starting size of position array: 1 element per 64 bytes of document
	// Tight fit when parsing small doc, only 0 to 2 reallocs when parsing large doc
	// Start at 8 when creating new document
	int nStartSize = m_strDoc.size() / 64 + 8;
	if ( (int)m_aPos.size() < nStartSize )
		m_aPos.resize( nStartSize );

	// Parse document
	bool bWellFormed = false;
	if ( m_strDoc.size() )
	{
		m_aPos[0].Clear();
		int iPos = x_ParseElem( 0 );
		if ( iPos > 0 )
		{
			m_aPos[0].iElemChild = iPos;
			bWellFormed = true;
		}
	}

	// Clear indexes if parse failed or empty document
	if ( ! bWellFormed )
	{
		m_aPos[0].Clear();
		m_iPosFree = 1;
	}

	ResetPos();
	return bWellFormed;
};

bool CMarkupSTL::IsWellFormed()
{
	if ( m_aPos.size() && m_aPos[0].iElemChild )
		return true;
	return false;
}

bool CMarkupSTL::Load( const char* szFileName )
{
	// Load document from file
	bool bResult = false;
	FILE* fp = fopen( szFileName, "rb" );
	if ( fp )
	{
		// Determine file length
		fseek( fp, 0L, SEEK_END );
		int nFileLen = ftell(fp);
		fseek( fp, 0L, SEEK_SET );

		// Load string
		allocator<char> mem;
		allocator<char>::pointer pBuffer = mem.allocate(nFileLen+1, NULL);
		if ( fread( pBuffer, nFileLen, 1, fp ) == 1 )
		{
			pBuffer[nFileLen] = '\0';
			bResult = SetDoc( pBuffer );
		}
		fclose(fp);
		mem.deallocate(pBuffer,1);
	}
	if ( ! bResult )
		SetDoc(NULL);
	return bResult;
}

bool CMarkupSTL::Save( const char* szFileName )
{
	// Save document to file
	bool bResult = false;
	FILE* fp = fopen( szFileName, "wb" );
	if ( fp )
	{
		// Save string
		int nFileLen = m_strDoc.size();
		if ( ! nFileLen )
			bResult = true;
		else if ( fwrite( m_strDoc.c_str(), nFileLen, 1, fp ) == 1 )
			bResult = true;
		fclose(fp);
	}
	return bResult;
}

bool CMarkupSTL::FindElem( const char* szName )
{
	// If szName is NULL or empty, go to next sibling element
	// Otherwise go to next sibling element with matching tag name
	// If the current position is valid, start looking from next
	// Change current position only if found
	//
	int iPos = m_iPos;
	if ( ! iPos )
	{
		if ( m_aPos.size() )
			iPos = m_aPos[0].iElemChild;
	}
	else
	{
		iPos = m_aPos[iPos].iElemNext;
	}

	while ( iPos )
	{
		// Compare tag name unless szName is not specified
		if ( szName == NULL || !szName[0] || x_STRMATCH(x_GetTagName(iPos),szName) )
		{
			// Assign new position
			m_iPos = iPos;
			m_iPosChild = 0;
			return true;
		}
		iPos = m_aPos[iPos].iElemNext;
	}
	return false;
}

bool CMarkupSTL::FindChildElem( const char* szName )
{
	// If szName is NULL or empty, go to next sibling child element
	// Otherwise go to next sibling child element with matching tag name
	// If the current child position is valid, start looking from next
	// Change current child position only if found
	//
	// Shorthand: call this with no current position means under root element
	if ( ! m_iPos )
		FindElem();

	// Is main position valid and not empty?
	if ( ! m_iPos || m_aPos[m_iPos].IsEmptyElement() )
		return false;

	// Is current child position valid?
	int iPosChild = m_iPosChild;
	if ( iPosChild )
		iPosChild = m_aPos[iPosChild].iElemNext;
	else
		iPosChild = m_aPos[m_iPos].iElemChild;

	// Search
	while ( iPosChild )
	{
		// Compare tag name unless szName is not specified
		if ( szName == NULL || !szName[0] || x_STRMATCH(x_GetTagName(iPosChild),szName) )
		{
			// Assign new position
			m_iPosChild = iPosChild;
			return true;
		}
		iPosChild = m_aPos[iPosChild].iElemNext;
	}
	return false;
}

bool CMarkupSTL::IntoElem()
{
	// Find child element unless there is already a child element position
	if ( ! m_iPosChild )
		FindChildElem();

	if ( m_iPosChild )
	{
		m_iPos = m_iPosChild;
		m_iPosChild = 0;
		++m_nLevel;
		return true;
	}

	return false;
}

bool CMarkupSTL::OutOfElem()
{
	// Go to parent element
	if ( m_iPos && m_aPos[m_iPos].iElemParent )
	{
		m_iPosChild = m_iPos;
		m_iPos = m_aPos[m_iPos].iElemParent;
		--m_nLevel;
		return true;
	}
	return false;
}

string CMarkupSTL::GetAttribName( int n ) const
{
	// Return nth attribute name of main position
	TokenPos token;
	token.nNext = m_aPos[m_iPos].nStartL + 1;
	for ( int nAttrib=0; nAttrib<=n; ++nAttrib )
		if ( ! x_FindAttrib(token) )
			return "";

	// Return substring of document
	return x_GetToken( token );
}

bool CMarkupSTL::SavePos( const char* szPosName )
{
	// Save current position in saved position map
	if ( m_iPos && szPosName )
	{
		SavedPos savedpos;
		savedpos.iPos = m_iPos;
		savedpos.iPosChild = m_iPosChild;
		savedpos.iLevel = m_nLevel;
		m_mapSavedPos[szPosName] = savedpos;
		return true;
	}
	return false;
}

bool CMarkupSTL::RestorePos( const char* szPosName )
{
	// Restore position if found in saved position map
	if ( szPosName )
	{
		mapSavedPosT::const_iterator iterSavePos = m_mapSavedPos.find( szPosName );
		if ( iterSavePos != m_mapSavedPos.end() )
		{
			SavedPos savedpos = (*iterSavePos).second;
			m_iPos = savedpos.iPos;
			m_iPosChild = savedpos.iPosChild;
			m_nLevel = savedpos.iLevel;
			return true;
		}
	}
	return false;
}

void*	CMarkupSTL::GetPos() const
{
	if ( m_iPos )
	{
		SavedPos* savedpos = new SavedPos;

		savedpos->iPos = m_iPos;
		savedpos->iPosChild = m_iPosChild;
		savedpos->iLevel = m_nLevel;

		return savedpos;
	}

	return NULL;
}

bool CMarkupSTL::SetPos( const void* savedpos )
{
	if ( !savedpos ) return false;

	m_iPos = ((SavedPos*)savedpos)->iPos;
	m_iPosChild = ((SavedPos*)savedpos)->iPosChild;
	m_nLevel = ((SavedPos*)savedpos)->iLevel;

	return true;
}

void	CMarkupSTL::ReleasePos( void* savedpos ) const
{
	if( !savedpos ) return;

	delete (SavedPos*) savedpos;
}

bool CMarkupSTL::GetOffsets( int& nStart, int& nEnd ) const
{
	// Return document offsets of current main position element
	// This is not part of EDOM but is used by the Markup project
	if ( m_iPos )
	{
		nStart = m_aPos[m_iPos].nStartL;
		nEnd = m_aPos[m_iPos].nEndR;
		return true;
	}
	return false;
}

bool CMarkupSTL::AddElem( const char* szName, const char* szValue )
{
	// Add an element after current main position
	int iPosParent = m_iPos? m_aPos[m_iPos].iElemParent : 0;
	m_iPosChild = 0;

	// Setting root element?
	if ( iPosParent == 0 )
	{
		if ( IsWellFormed() )
			return false;

		m_strDoc.erase();
	}

	m_iPos = x_Add( iPosParent, m_iPos, szName, szValue );
	return true;
}

bool CMarkupSTL::AddChildElem( const char* szName, const char* szValue )
{
	// Add a child element under main position, after current child position
	if ( ! m_iPos )
		return false;

	// If no child position, add after last sibling
	int iPosLast = m_aPos[m_iPos].iElemChild;
	if ( ! m_iPosChild && iPosLast )
	{
		m_iPosChild = iPosLast;
		while ( (iPosLast=m_aPos[iPosLast].iElemNext) != 0 )
			m_iPosChild = iPosLast;
	}

	m_iPosChild = x_Add( m_iPos, m_iPosChild, szName, szValue );
	return true;
}

bool CMarkupSTL::InsertChildElem( const char* szName, const char* szValue )
{
	// Insert a child element under main position, before current child position
	if ( ! m_iPos )
		return false;

	// Get first child of parent
	int iPosPrev = m_aPos[m_iPos].iElemChild;

	// Is current position clear or first sibling?
	if ( ! m_iPosChild || iPosPrev == m_iPosChild )
	{
		// Add as first child under parent
		m_iPosChild = x_Add( m_iPos, 0, szName, szValue );
		return true;
	}

	// Find position before current position
	while ( m_aPos[iPosPrev].iElemNext != m_iPosChild )
		iPosPrev = m_aPos[iPosPrev].iElemNext;

	// Add after previous position
	m_iPosChild = x_Add( m_iPos, iPosPrev, szName, szValue );
	return true;
}

bool CMarkupSTL::AddAttrib( const char* szAttrib, int nValue )
{
	// Convert integer to string and call AddAttrib
	char szVal[25];
	itoa( nValue, szVal, 10 );
	return AddAttrib( szAttrib, szVal );
}

bool CMarkupSTL::AddAttrib( const char* szAttrib, const char* szValue )
{
	// Add attribute to current main position element
	if ( m_iPos )
	{
		x_AddAttrib( m_iPos, szAttrib, szValue );
		return true;
	}
	return false;
}

bool CMarkupSTL::AddChildAttrib( const char* szAttrib, int nValue )
{
	// Convert integer to string and call AddChildAttrib
	char szVal[25];
	itoa( nValue, szVal, 10 );
	return AddChildAttrib( szAttrib, szVal );
}

bool CMarkupSTL::AddChildAttrib( const char* szAttrib, const char* szValue )
{
	// Add attribute to current child position element
	if ( m_iPosChild )
	{
		x_AddAttrib( m_iPosChild, szAttrib, szValue );
		return true;
	}
	return false;
}

bool CMarkupSTL::AddChildSubDoc( const char* szSubDoc )
{
	// Add a subdocument under main position, after current child position
	if ( ! m_iPos )
		return false;

	// If no child position, add after last sibling
	int iPosLast = m_aPos[m_iPos].iElemChild;
	int iPosBefore = m_iPosChild;
	if ( ! iPosBefore && iPosLast )
	{
		iPosBefore = iPosLast;
		while ( (iPosLast=m_aPos[iPosLast].iElemNext) != 0 )
			iPosBefore = iPosLast;
	}

	int iPosSubDoc = x_AddSubDoc( m_iPos, iPosBefore, szSubDoc );
	if ( ! iPosSubDoc )
		return false;

	m_iPosChild = iPosSubDoc;
	return true;
}

bool CMarkupSTL::InsertChildSubDoc( const char* szSubDoc )
{
	// Add a subdocument under main position, after current child position
	if ( ! m_iPos )
		return false;

	// Get first child of parent
	int iPosPrev = m_aPos[m_iPos].iElemChild;

	// Is current child position after first child?
	if ( m_iPosChild && iPosPrev != m_iPosChild )
	{
		// Find position before current position
		while ( m_aPos[iPosPrev].iElemNext != m_iPosChild )
			iPosPrev = m_aPos[iPosPrev].iElemNext;
	}
	else
	{
		iPosPrev = 0;
	}

	// Add after previous position
	int iPosSubDoc = x_AddSubDoc( m_iPos, iPosPrev, szSubDoc );
	if ( ! iPosSubDoc )
		return false;

	m_iPosChild = iPosSubDoc;
	return true;
}

string CMarkupSTL::GetChildSubDoc() const
{
	if ( m_iPosChild )
	{
		int nL = m_aPos[m_iPosChild].nStartL;
		int nR = m_aPos[m_iPosChild].nNext;
		return m_strDoc.substr( nL, nR - nL );
	}
	return "";
}

bool CMarkupSTL::RemoveElem()
{
	// Remove current main position element
	if ( m_iPos )
	{
		m_iPosChild = 0;
		m_iPos = x_Remove( m_iPos );
		return true;
	}
	return false;
}

bool CMarkupSTL::RemoveChildElem()
{
	// Remove current child position element
	if ( m_iPosChild )
	{
		m_iPosChild = x_Remove( m_iPosChild );
		return true;
	}
	return false;
}

bool CMarkupSTL::SetAttrib( const char* szAttrib, int nValue )
{
	// Convert integer to string and call SetAttrib
	char szVal[25];
	itoa( nValue, szVal, 10 );
	return SetAttrib( szAttrib, szVal );
}

bool CMarkupSTL::SetAttrib( const char* szAttrib, const char* szValue )
{
	// Set attribute in current main position element
	if ( m_iPos )
	{
		x_SetAttrib( m_iPos, szAttrib, szValue );
		return true;
	}
	return false;
}

bool CMarkupSTL::SetChildAttrib( const char* szAttrib, int nValue )
{
	// Convert integer to string and call SetChildAttrib
	char szVal[25];
	itoa( nValue, szVal, 10 );
	return SetChildAttrib( szAttrib, szVal );
}

bool CMarkupSTL::SetChildAttrib( const char* szAttrib, const char* szValue )
{
	// Set attribute in current child position element
	if ( m_iPosChild )
	{
		x_SetAttrib( m_iPosChild, szAttrib, szValue );
		return true;
	}
	return false;
}

//////////////////////////////////////////////////////////////////////
// Private Methods
//////////////////////////////////////////////////////////////////////

int CMarkupSTL::x_GetFreePos()
{
	//
	// This returns the index of the next unused ElemPos in the array
	//
	if ( m_iPosFree == m_aPos.size() )
		m_aPos.resize( m_iPosFree + m_iPosFree / 2 );
	++m_iPosFree;
	return m_iPosFree - 1;
}

int CMarkupSTL::x_ReleasePos()
{
	//
	// This decrements the index of the next unused ElemPos in the array
	// allowing the element index returned by GetFreePos() to be reused
	//
	--m_iPosFree;
	return 0;
}

int CMarkupSTL::x_ParseError( const char* szError, const char* szTag )
{
	if ( szTag )
	{
		char szFormat[300];
		sprintf( szFormat, szError, szTag );
		m_strError = szFormat;
	}
	else
		m_strError = szError;
	x_ReleasePos();
	return -1;
}

int CMarkupSTL::x_ParseElem( int iPosParent )
{
	// This is either called by SetDoc, x_AddSubDoc, or itself recursively
	// m_aPos[iPosParent].nEndL is where to start parsing for the child element
	// This returns the new position if a tag is found, otherwise zero
	// In all cases we need to get a new ElemPos, but release it if unused
	//
	int iPos = x_GetFreePos();
	m_aPos[iPos].nStartL = m_aPos[iPosParent].nEndL;
	m_aPos[iPos].nNext = m_aPos[iPosParent].nStartR + 1;
	m_aPos[iPos].iElemParent = iPosParent;
	m_aPos[iPos].iElemChild = 0;
	m_aPos[iPos].iElemNext = 0;

	// Start Tag
	// A loop is used to ignore all remarks tags and special tags
	// i.e. <?xml version="1.0"?>, and <!-- comment here -->
	// So any tag beginning with ? or ! is ignored
	// Loop past ignored tags
	TokenPos token;
	token.nNext = m_aPos[iPosParent].nEndL;
	string strName;
	while ( strName.empty() )
	{
		// Look for left angle bracket of opening tag
		m_aPos[iPos].nStartL = token.nNext;
		if ( ! x_FindChar( m_aPos[iPos].nStartL, '<' ) )
			return x_ParseError( "Element tag not found" );

		// Set parent's End tag to start looking from here (or later)
		m_aPos[iPosParent].nEndL = m_aPos[iPos].nStartL;

		// Determine whether this is an element, comment or version tag
		const char* szEndOfTag = NULL;
		token.nNext = m_aPos[iPos].nStartL + 1;
		if ( x_FindToken( token ) )
		{
			if ( token.bIsString )
				return x_ParseError( "Tag starts with quote" );
			char cFirstChar = m_strDoc[token.nL];
			if ( cFirstChar == '?' )
				szEndOfTag = "?>"; // version
			else if ( cFirstChar == '!' )
			{
				char cSecondChar = 0;
				if ( token.nL+1 < (int)m_strDoc.length() )
					cSecondChar = m_strDoc[token.nL+1];
				if ( cSecondChar == '[' )
					szEndOfTag = "]]>"; // CDATA section
				else if ( cSecondChar == '-' )
					szEndOfTag = "-->"; // comment
				else
					szEndOfTag = ">"; // DTD
			}
			else if ( cFirstChar != '/' )
			{
				strName = x_GetToken( token );
				szEndOfTag = ">";
			}
			else
				return x_ReleasePos(); // probably end tag of parent
		}
		else
			return x_ParseError( "Abrupt end within tag" );

		// Look for end of tag
		token.nNext = m_strDoc.find( szEndOfTag, token.nNext );
		if ( token.nNext == string::npos )
			return x_ParseError( "End of tag not found" );
	}
	m_aPos[iPos].nStartR = token.nNext;

	// Is ending mark within opening tag, i.e. empty element?
	if ( m_strDoc[m_aPos[iPos].nStartR-1] == '/' )
	{
		// Empty element
		// Close tag left is set to ending mark, and right to open tag right
		m_aPos[iPos].nEndL = m_aPos[iPos].nStartR-1;
		m_aPos[iPos].nEndR = m_aPos[iPos].nStartR;
	}
	else // look for end tag
	{
		// Element probably has contents
		// Determine where to start looking for left angle bracket of end tag
		// This is done by recursively parsing the contents of this element
		int iInner, iInnerPrev = 0;
		m_aPos[iPos].nEndL = m_aPos[iPos].nStartR + 1;
		while ( (iInner=x_ParseElem(iPos)) > 0 )
		{
			// Set links to iInner
			if ( iInnerPrev )
				m_aPos[iInnerPrev].iElemNext = iInner;
			else
				m_aPos[iPos].iElemChild = iInner;
			iInnerPrev = iInner;

			// Set offset to reflect child
			m_aPos[iPos].nEndL = m_aPos[iInner].nEndR + 1;
		}
		if ( iInner == -1 )
			return -1;

		// Look for left angle bracket of end tag
		if ( ! x_FindChar( m_aPos[iPos].nEndL, '<' ) )
			return x_ParseError( "End tag of %s element not found", strName.c_str() );

		// Look through tokens of end tag
		token.nNext = m_aPos[iPos].nEndL + 1;
		int nTokenCount = 0;
		while ( x_FindToken( token ) )
		{
			++nTokenCount;
			if ( ! token.bIsString )
			{
				// Is first token not an end slash mark?
				if ( nTokenCount == 1 && m_strDoc[token.nL] != '/' )
					return x_ParseError( "Expecting end tag of element %s", strName.c_str() );

				else if ( nTokenCount == 2 && strName != x_GetToken( token ) )
					return x_ParseError( "End tag does not correspond to %s", strName.c_str() );

				// Else is it a right angle bracket?
				else if ( m_strDoc[token.nL] == '>' )
					break;
			}
		}

		// Was a right angle bracket not found?
		if ( ! token.IsValid() || nTokenCount < 2 )
			return x_ParseError( "End tag not completed for element %s", strName.c_str() );
		m_aPos[iPos].nEndR = token.nL;
	}

	// Successfully found positions of angle brackets
	m_aPos[iPos].nNext = m_aPos[iPos].nEndR;
	x_FindChar( m_aPos[iPos].nNext, '<' );
	return iPos;
}

bool CMarkupSTL::x_FindChar( int&n, char c ) const
{
	// Look for char c starting at n, and set n to point to it
	// Return false if not found before end of document
	int nMax = m_strDoc.size();
	while ( n < nMax && m_strDoc[n] != c )
		++n;
	if ( n == nMax )
		return false;
	return true;
}

bool CMarkupSTL::x_FindToken( CMarkupSTL::TokenPos& token ) const
{
	// Starting at token.nNext, find the next token
	// upon successful return, token.nNext points after the retrieved token
	int nMax = m_strDoc.size();
	int n = token.nNext;

	// Statically defined strings for whitespace and special chars
	static string strWhitespace = " \t\n\r";
	static string strSpecial = "<>=\\/?!";

	// By-pass leading whitespace
	while ( n < nMax && strWhitespace.find(m_strDoc[n]) != string::npos )
		++n;

	// Are we still within the document?
	token.bIsString = false;
	if ( n < nMax )
	{
		// Is it an opening quote?
		if ( m_strDoc[n] == '\"' )
		{
			// Move past opening quote
			++n;
			token.nL = n;

			// Look for closing quote
			x_FindChar( n, '\"' );

			// Set right to before closing quote
			token.nR = n-1;

			// Set n past closing quote unless at end of document
			if ( n < nMax )
				++n;

			// Set flag
			token.bIsString = true;
		}
		else
		{
			// Go until special char or whitespace
			token.nL = n;
			while ( n < nMax &&
				strSpecial.find(m_strDoc[n]) == string::npos &&
				strWhitespace.find(m_strDoc[n]) == string::npos
				)
				++n;

			// Adjust end position if it is one special char
			if ( n == token.nL )
				++n; // it is a special char
			token.nR = n-1;
		}
	}

	token.nNext = n;
	if ( n >= nMax )
		return false;

	// nEnd points to one past last char of token
	return true;
}

string CMarkupSTL::x_GetToken( const CMarkupSTL::TokenPos& token ) const
{
	// The token contains indexes into the document identifying a small substring
	// Build the substring from those indexes and return it
	if ( ! token.IsValid() )
		return "";
	return m_strDoc.substr( token.nL,
		token.nR - token.nL + ((token.nR<(int)m_strDoc.size())? 1:0) );
}

string CMarkupSTL::x_GetTagName( int iPos ) const
{
	// Return the tag name at specified element
	TokenPos token;
	token.nNext = m_aPos[iPos].nStartL + 1;
	if ( ! iPos || ! x_FindToken( token ) )
		return "";

	// Return substring of document
	return x_GetToken( token );
}

bool CMarkupSTL::x_FindAttrib( CMarkupSTL::TokenPos& token, const char* szAttrib ) const
{
	// If szAttrib is NULL find next attrib, otherwise find named attrib
	// Return true if found
	int nAttrib = 0;
	for ( int nCount = 0; x_FindToken(token); ++nCount )
	{
		if ( ! token.bIsString )
		{
			// Is it the right angle bracket?
			if ( m_strDoc[token.nL] == '>' || m_strDoc[token.nL] == '/' )
				break; // attrib not found

			// Equal sign
			if ( m_strDoc[token.nL] == '=' )
				continue;

			// Potential attribute
			if ( ! nAttrib && nCount )
			{
				// Attribute name search?
				if ( ! szAttrib || ! szAttrib[0] )
					return true; // return with token at attrib name

				if ( x_STRMATCH(x_GetToken(token),szAttrib) )
					nAttrib = nCount;
			}
		}
		else if ( nAttrib && nCount == nAttrib + 2 )
		{
			return true;
		}
	}

	// Not found
	return false;
}

string CMarkupSTL::x_GetAttrib( int iPos, const char* szAttrib ) const
{
	// Return the value of the attrib at specified element
	TokenPos token;
	token.nNext = m_aPos[iPos].nStartL + 1;
	if ( szAttrib && x_FindAttrib( token, szAttrib ) )
		return x_TextFromDoc( token.nL, token.nR - ((token.nR<(int)m_strDoc.size())?0:1) );
	return "";
}

int CMarkupSTL::x_SetAttrib( int iPos, const char* szAttrib, const char* szValue )
{
	// Set attribute in iPos element
	string strInsert;
	if ( iPos )
	{
		TokenPos token;
		token.nNext = m_aPos[iPos].nStartL + 1;
		int nL, nReplace = 0;
		if ( x_FindAttrib( token, szAttrib ) )
		{
			// Decision: for empty value leaving attrib="" instead of removing attrib
			// Replace value only
			strInsert = x_TextToDoc( szValue, true );
			nL = token.nL;
			nReplace = token.nR-token.nL+1;
		}
		else
		{
			// Insert string name value pair
			string strValue = x_TextToDoc( szValue, true );
			strInsert = " ";
			strInsert.append(szAttrib);
			strInsert.append("=\"");
			strInsert.append(strValue);
			strInsert.append("\"");
			// take into account whether it is an empty element
			nL = m_aPos[iPos].nStartR - (m_aPos[iPos].IsEmptyElement()?1:0);
		}

		x_DocChange( nL, nReplace, strInsert );
		int nAdjust = strInsert.size() - nReplace;
		m_aPos[iPos].nStartR += nAdjust;
		m_aPos[iPos].nEndL += nAdjust;
		m_aPos[iPos].nEndR += nAdjust;
		m_aPos[iPos].nNext += nAdjust;
		x_Adjust( iPos, nAdjust );
	}
	return strInsert.size();
}

bool CMarkupSTL::x_SetData( int iPos, const char* szData, int nCDATA )
{
	// Set data in iPos element
	ElemPos& ep = m_aPos[iPos];
	if ( ! iPos || ep.iElemChild )
		return false;

	// Build csInsert from szData based on nCDATA
	string strInsert;
	if ( nCDATA != 0 )
	{
		// To protect well-formed document,
		// allow nCDATA to be unset if data contains "]]>"
		if ( strstr(szData,"]]>") != NULL )
			nCDATA = 0;
		else
		{
			strInsert = "<![CDATA[]]>";
			strInsert.insert( 9, szData );
		}
	}
	if ( nCDATA == 0 )
		strInsert = x_TextToDoc( szData );

	// Decide where to insert
	int nL, nReplace;
	if ( ep.IsEmptyElement() )
	{
		nL = ep.nEndL;
		nReplace = 1;

		// Pre-adjust since <NAME/> becomes <NAME>data</NAME>
		string strTagName = x_GetTagName( iPos );
		ep.nStartR -= 1;
		ep.nEndL -= (1 + strTagName.size());
		string strFormat;
		strFormat = ">";
		strFormat.append(strInsert);
		strFormat.append("</");
		strFormat.append(strTagName);
		strInsert = strFormat;
	}
	else
	{
		nL = ep.nStartR+1;
		nReplace = ep.nEndL - ep.nStartR - 1;
	}
	x_DocChange( nL, nReplace, strInsert );
	int nAdjust = strInsert.size() - nReplace;
	x_Adjust( iPos, nAdjust );
	ep.nEndL += nAdjust;
	ep.nEndR += nAdjust;
	ep.nNext += nAdjust;
	return true;
}

string CMarkupSTL::x_GetData( int iPos ) const
{
	// Return a string representing data between start and end tag
	// Return empty string if there are any children elements
	if ( ! m_aPos[iPos].iElemChild && ! m_aPos[iPos].IsEmptyElement() )
	{
		// See if it is a CDATA section
		TokenPos token;
		token.nNext = m_aPos[iPos].nStartR+1;
		if ( x_FindToken( token ) && m_strDoc[token.nL] == '<'
				&& token.nL + 11 < m_aPos[iPos].nEndL
				&& m_strDoc.compare(token.nL+1,8,"![CDATA[") == 0 )
		{
			int nEndCDATA = m_strDoc.find( "]]>", token.nNext );
			if ( nEndCDATA != string::npos && nEndCDATA < m_aPos[iPos].nEndL )
			{
				return m_strDoc.substr( token.nL+9, nEndCDATA-token.nL-9 );
			}
		}
		return x_TextFromDoc( m_aPos[iPos].nStartR+1, m_aPos[iPos].nEndL-1 );
	}
	return "";
}

string CMarkupSTL::x_TextToDoc( const char* szText, bool bAttrib ) const
{
	//
	// &lt;   less than
	// &amp;  ampersand
	// &gt;   greater than
	//
	// and for attributes:
	//
	// &apos; apostrophe or single quote
	// &quot; double quote
	//
	static char* szaReplace[] = { "&lt;","&amp;","&gt;","&apos;","&quot;" };
	const char* pFind = bAttrib?"<&>\'\"":"<&>";
	const char* pSource = szText;
	string strResult;
	int nLen = strlen( szText );
	strResult.reserve( nLen + nLen / 10 );
	char cSource = *pSource;
	char* pFound;
	while ( cSource )
	{
		if ( (pFound=strchr(pFind,cSource)) != NULL )
		{
			pFound = szaReplace[pFound-pFind];
			strResult.append(pFound);
		}
		else
		{
			strResult += cSource;
		}
		cSource = *(++pSource);
	}
	return strResult;
}

string CMarkupSTL::x_TextFromDoc( int nLeft, int nRight ) const
{
	//
	// Conveniently the result is always the same or shorter in length
	//
	static char* szaCode[] = { "lt;","amp;","gt;","apos;","quot;" };
	static int anCodeLen[] = { 3,4,3,5,5 };
	static char* szSymbol = "<&>\'\"";
	string strResult;
	strResult.reserve( nRight - nLeft + 1 );
	const char* pSource = m_strDoc.c_str();
	int nChar = nLeft;
	char cSource = pSource[nChar];
	while ( nChar <= nRight )
	{
		if ( cSource == '&' )
		{
			// If no match is found it means XML doc is invalid
			// no devastating harm done, ampersand code will just be left in result
			for ( int nMatch = 0; nMatch < 5; ++nMatch )
			{
				if ( nChar <= nRight - anCodeLen[nMatch]
					&& strncmp(szaCode[nMatch],&pSource[nChar+1],anCodeLen[nMatch]) == 0 )
				{
					cSource = szSymbol[nMatch];
					nChar += anCodeLen[nMatch];
					break;
				}
			}
		}
		strResult += cSource;
		nChar++;
		cSource = pSource[nChar];
	}
	return strResult;
}

void CMarkupSTL::x_DocChange( int nLeft, int nReplace, string& strInsert )
{
	// Insert strInsert int m_strDoc at nLeft replacing nReplace chars
	//
	m_strDoc.replace( nLeft, nReplace, strInsert);
}

void CMarkupSTL::x_Adjust( int iPos, int nShift )
{
	// Loop through affected elements and adjust indexes
	// Does not affect iPos itself
	// Algorithm:
	// 1. update next siblings and all their children
	// 2. then go up a level update end points and to step 1
	int iPosTop = m_aPos[iPos].iElemParent;
	while ( iPos )
	{
		// Were we at containing parent of affected position?
		bool bPosTop = false;
		if ( iPos == iPosTop )
		{
			// Move iPosTop up one towards root
			iPosTop = m_aPos[iPos].iElemParent;
			bPosTop = true;
		}

		// Traverse to the next update position
		if ( ! bPosTop && m_aPos[iPos].iElemChild )
		{
			// Depth first
			iPos = m_aPos[iPos].iElemChild;
		}
		else if ( m_aPos[iPos].iElemNext )
		{
			iPos = m_aPos[iPos].iElemNext;
		}
		else
		{
			// Look for next sibling of a parent of iPos
			// When going back up, parents have already been done except iPosTop
			while ( (iPos=m_aPos[iPos].iElemParent) != 0 && iPos != iPosTop )
				if ( m_aPos[iPos].iElemNext )
				{
					iPos = m_aPos[iPos].iElemNext;
					break;
				}
		}

		// Shift indexes at iPos
		if ( iPos != iPosTop )
		{
			// Move the start tag indexes
			// Don't do this for containing parent tag
			m_aPos[iPos].nStartL += nShift;
			m_aPos[iPos].nStartR += nShift;
		}
		// Move end tag indexes
		m_aPos[iPos].nEndL += nShift;
		m_aPos[iPos].nEndR += nShift;
		m_aPos[iPos].nNext += nShift;
	}
}

int CMarkupSTL::x_Add( int iPosParent, int iPosBefore, const char* szName, const char* szValue )
{
	// Create element and modify positions of affected elements
	// if iPosBefore is NULL, insert as first element under parent
	// If no szValue is specified, an empty element is created
	// i.e. either <NAME>value</NAME> or <NAME/>
	//
	int iPos = x_GetFreePos();
	bool bEmptyParent = false;
	if ( iPosBefore )
	{
		// Follow iPosBefore
		m_aPos[iPos].nStartL = m_aPos[iPosBefore].nNext;
	}
	else if ( m_aPos[iPosParent].iElemChild )
	{
		// Insert before first child of parent
		m_aPos[iPos].nStartL = m_aPos[m_aPos[iPosParent].iElemChild].nStartL;
	}
	else if ( m_aPos[iPosParent].IsEmptyElement() )
	{
		// Parent has no separate end tag
		m_aPos[iPos].nStartL = m_aPos[iPosParent].nStartR + 2;
		bEmptyParent = true;
	}
	else
	{
		// Parent has content, but no children
		m_aPos[iPos].nStartL = m_aPos[iPosParent].nEndL;
	}

	// Set links
	m_aPos[iPos].iElemParent = iPosParent;
	m_aPos[iPos].iElemChild = 0;
	if ( iPosBefore )
	{
		m_aPos[iPos].iElemNext = m_aPos[iPosBefore].iElemNext;
		m_aPos[iPosBefore].iElemNext = iPos;
	}
	else
	{
		m_aPos[iPos].iElemNext = m_aPos[iPosParent].iElemChild;
		m_aPos[iPosParent].iElemChild = iPos;
	}

	// Create string for insert
	string strInsert;
	int nLenName = strlen(szName);
	int nLenValue = szValue? strlen(szValue) : 0;
	if ( ! nLenValue )
	{
		// <NAME/> empty element
		strInsert = "<";
		strInsert.append(szName);
		strInsert.append("/>\r\n");
		m_aPos[iPos].nStartR = m_aPos[iPos].nStartL + nLenName + 2;
		m_aPos[iPos].nEndL = m_aPos[iPos].nStartR - 1;
		m_aPos[iPos].nEndR = m_aPos[iPos].nEndL + 1;
		m_aPos[iPos].nNext = m_aPos[iPos].nEndR + 3;
	}
	else
	{
		// <NAME>value</NAME>
		string strValue = x_TextToDoc( szValue );
		nLenValue = strValue.size();
		strInsert = "<";
		strInsert.append(szName);
		strInsert.append(">");
		strInsert.append(strValue);
		strInsert.append("</");
		strInsert.append(szName);
		strInsert.append(">\r\n");
		m_aPos[iPos].nStartR = m_aPos[iPos].nStartL + nLenName + 1;
		m_aPos[iPos].nEndL = m_aPos[iPos].nStartR + nLenValue + 1;
		m_aPos[iPos].nEndR = m_aPos[iPos].nEndL + nLenName + 2;
		m_aPos[iPos].nNext = m_aPos[iPos].nEndR + 3;
	}

	// Insert
	int nReplace = 0, nLeft = m_aPos[iPos].nStartL;
	if ( bEmptyParent )
	{
		string strParentTagName = x_GetTagName(iPosParent);
		string strFormat;
		strFormat = ">\r\n";
		strFormat.append(strInsert);
		strFormat.append("</");
		strFormat.append(strParentTagName);
		strInsert = strFormat;
		nLeft -= 3;
		nReplace = 1;
		// x_Adjust is going to update all affected indexes by one amount
		// This will satisfy all except the empty parent
		// Here we pre-adjust for the empty parent
		// The empty tag slash is removed
		m_aPos[iPosParent].nStartR -= 1;
		// For the newly created end tag, see the following example:
		// <A/> (len 4) becomes <A><B/></A> (len 11)
		// In x_Adjust everything will be adjusted 11 - 4 = 7
		// But the nEndL of element A should only be adjusted 5
		m_aPos[iPosParent].nEndL -= (strParentTagName.size() + 1);
	}
	x_DocChange( nLeft, nReplace, strInsert );
	x_Adjust( iPos, strInsert.size() - nReplace );

	// Return the index of the new element
	return iPos;
}

int CMarkupSTL::x_AddSubDoc( int iPosParent, int iPosBefore, const char* szSubDoc )
{
	// Add subdocument, parse, and modify positions of affected elements
	// if iPosBefore is NULL, insert as first element under parent
	//
	int nParentEndLBeforeAdd = m_aPos[iPosParent].nEndL;
	int iPosFreeBeforeAdd = m_iPosFree;

	// Subdocument may start with a version tag i.e. <?xml version="1.0"?>
	// To skip version tag of subdocument, use #defined parse routine
	// SUBDOCFIND looks for a char in the subdocument, returns 0 (fail) if not found
	#define SUBDOCFIND(c) {while(*szSD && *szSD!=c) ++szSD; if(*szSD!=c) return 0;}
	const char* szSD = szSubDoc;
	SUBDOCFIND('<')
	if ( szSD[1] == '?' )
	{
		SUBDOCFIND('>')
		--szSD; // back up one char to look for ?
		if ( *szSD == '?' )
			SUBDOCFIND('<')
		else
			szSD = szSubDoc;
	}
	string strInsert = szSD;

	// Determine where to insert
	bool bEmptyParent = false;
	if ( iPosBefore )
	{
		// Follow iPosBefore
		m_aPos[iPosParent].nEndL = m_aPos[iPosBefore].nNext;
	}
	else if ( m_aPos[iPosParent].iElemChild )
	{
		// Insert before first child of parent
		m_aPos[iPosParent].nEndL = m_aPos[m_aPos[iPosParent].iElemChild].nStartL;
	}
	else if ( m_aPos[iPosParent].IsEmptyElement() )
	{
		// Parent has no separate end tag
		bEmptyParent = true;
	}

	// Insert subdocument
	int nReplace = 0, nLeft = m_aPos[iPosParent].nEndL;
	string strParentTagName;
	if ( bEmptyParent )
	{
		strParentTagName = x_GetTagName(iPosParent);
		string strFormat;
		strFormat = ">\r\n";
		strFormat.append(strInsert);
		strFormat.append("</");
		strFormat.append(strParentTagName);
		strInsert = strFormat;
		m_aPos[iPosParent].nEndL = m_aPos[iPosParent].nStartR + 2;
		nLeft = m_aPos[iPosParent].nStartR - 1;
		nReplace = 1;
	}
	x_DocChange( nLeft, nReplace, strInsert );

	// Parse subdocument
	int iPos = x_ParseElem(iPosParent);
	m_aPos[iPosParent].nEndL = nParentEndLBeforeAdd;
	if ( iPos <= 0 )
	{
		// Abort because not well-formed
		string strRevert = bEmptyParent?"/":"";
		x_DocChange( nLeft, strInsert.size(), strRevert );
		m_iPosFree = iPosFreeBeforeAdd;
		iPos = 0;
	}
	else
	{
		// Link in parent and siblings
		m_aPos[iPos].iElemParent = iPosParent;
		if ( iPosBefore )
		{
			m_aPos[iPos].iElemNext = m_aPos[iPosBefore].iElemNext;
			m_aPos[iPosBefore].iElemNext = iPos;
		}
		else
		{
			m_aPos[iPos].iElemNext = m_aPos[iPosParent].iElemChild;
			m_aPos[iPosParent].iElemChild = iPos;
		}

		// Make empty parent pre-adjustment
		if ( bEmptyParent )
		{
			m_aPos[iPosParent].nStartR -= 1;
			m_aPos[iPosParent].nEndL -= (strParentTagName.size() + 1);
		}

		// Adjust, but don't adjust children of iPos
		int iElemChild = m_aPos[iPos].iElemChild;
		m_aPos[iPos].iElemChild = 0;
		x_Adjust( iPos, strInsert.size() - nReplace );
		m_aPos[iPos].iElemChild = iElemChild;
	}

	// Return the index of the new element
	return iPos;
}

int CMarkupSTL::x_AddAttrib( int iPos, const char* szAttrib, const char* szValue )
{
	// Add attribute to iPos element
	string strInsert;
	if ( iPos )
	{
		// Insert string taking into account whether it is a single tag
		string strValue = x_TextToDoc( szValue, true );
		strInsert = " ";
		strInsert.append(szAttrib);
		strInsert.append("=\"");
		strInsert.append(strValue);
		strInsert.append("\"");
		int nL = m_aPos[iPos].nStartR - (m_aPos[iPos].IsEmptyElement()?1:0);
		x_DocChange( nL, 0, strInsert );

		int nLen = strInsert.size();
		m_aPos[iPos].nStartR += nLen;
		m_aPos[iPos].nEndL += nLen;
		m_aPos[iPos].nEndR += nLen;
		m_aPos[iPos].nNext += nLen;
		x_Adjust( iPos, nLen );
	}
	return strInsert.size();
}

int CMarkupSTL::x_Remove( int iPos )
{
	// Remove element and all contained elements
	// Return new position
	//
	int iPosParent = m_aPos[iPos].iElemParent;

	// Find previous sibling and bypass removed element
	// This leaves orphan positions in m_aPos array
	int iPosLook = m_aPos[iPosParent].iElemChild;
	int iPosPrev = 0;
	while ( iPosLook != iPos )
	{
		iPosPrev = iPosLook;
		iPosLook = m_aPos[iPosLook].iElemNext;
	}
	if ( iPosPrev )
		m_aPos[iPosPrev].iElemNext = m_aPos[iPos].iElemNext;
	else
		m_aPos[iPosParent].iElemChild = m_aPos[iPos].iElemNext;

	// Remove from document
	// Links have been changed to go around removed element
	// But element position and links are still valid
	int nLen = m_aPos[iPos].nNext - m_aPos[iPos].nStartL;
	x_DocChange( m_aPos[iPos].nStartL, nLen, string() );
	x_Adjust( iPos, 0 - nLen );
	return iPosPrev;
}

