#pragma once


//
// This class handles a frame range for a state change: if the current frame played
// falls into the range and the current state id is changed to the state id
// given by the CTRStateChange class, then the current animation is chained to
// the animation given by the next anim / frame fields of this class
//
class CTRAnimDispatch
{
public:

    // Constructors / destructors
    CTRAnimDispatch() : m_dwStartFrame(0), m_dwEndFrame(0) ,m_dwNextAnimation(0), m_dwNextFrame(0) { }
    CTRAnimDispatch(DWORD dwStartFrame, DWORD dwEndFrame, DWORD dwNextAnimation, DWORD dwNextFrame)
    {
        vSetFrameRange(dwStartFrame, dwEndFrame);
        vSetNextAnim(dwNextAnimation, dwNextFrame);
    }

    // Accessors
    void            vGetFrameRange(DWORD& dwStartFrame, DWORD& dwEndFrame)   const  { dwStartFrame = m_dwStartFrame; dwEndFrame = m_dwEndFrame; }
    void            vGetNextAnim(DWORD& dwNextAnimation, DWORD& dwNextFrame) const  { dwNextAnimation = m_dwNextAnimation; dwNextFrame = m_dwNextFrame; }

    void            vSetFrameRange(DWORD dwStartFrame, DWORD dwEndFrame)            { ASSERT(m_dwStartFrame <= m_dwEndFrame); m_dwStartFrame = dwStartFrame; m_dwEndFrame = dwEndFrame; }
    void            vSetNextAnim(DWORD dwNextAnimation, DWORD dwNextFrame)          { m_dwNextAnimation = dwNextAnimation; m_dwNextFrame = dwNextFrame; }

private:

    // Data members
    DWORD                   m_dwStartFrame;             // start frame of the range
    DWORD                   m_dwEndFrame;               // end frame of the range
    DWORD                   m_dwNextAnimation;          // next animation index to chain to
    DWORD                   m_dwNextFrame;              // next frame in the next animation
};


//
// This class handles the state changes, which allow to chain to another
// animation when the state changes during the playing of an animation
//
class CTRStateChange
{
public:

    // Constructors / destructors
    CTRStateChange(DWORD dwStateId) : m_dwStateId(dwStateId), m_LstAnimDispatch(), m_dwTagCounter(0)
                                                        { }
    virtual ~CTRStateChange()                           { m_LstAnimDispatch.clear(); }

    // Accessors
    DWORD     dwGetStateId()                    const   { return m_dwStateId; }
    DWORD     dwGetNumAnimDispatch()            const   { return m_LstAnimDispatch.size(); }
    const CTRAnimDispatch&  GetAnimDispatch(DWORD dwKey, DWORD& dwOtherKey, bool boKeyIsTag) const;

    // Operations
    const void  vPushAnimDispatch(const CTRAnimDispatch& AnimDispatch)
                                                        { m_LstAnimDispatch.push_back( OneAnimDispatch(AnimDispatch, m_dwTagCounter++) ); }
    void        vPackAnimDispatches()                   { m_LstAnimDispatch.resize(m_LstAnimDispatch.size()); }
    const void  vRemoveAnimDispatch( const DWORD dwTag );

private:

    DWORD       m_dwTagCounter;

    class OneAnimDispatch
    {
    public:

        OneAnimDispatch( const CTRAnimDispatch& AnimDispatch, DWORD dwTag ) : m_AnimDispatch( AnimDispatch ), m_dwTag( dwTag ) { }
        OneAnimDispatch() : m_dwTag(0) { }

        CTRAnimDispatch m_AnimDispatch;
        DWORD           m_dwTag;
    };

    typedef std::vector<OneAnimDispatch> TRLstAnimDispatch;

    // Data members
    DWORD                   m_dwStateId;                // state id
    TRLstAnimDispatch       m_LstAnimDispatch;          // list of anim dispatches. If the current state changes to
                                                        // 'state id' and if the current frame falls into one of the
                                                        // frame range of the anim dispatch list, then we chain to the
                                                        // next animation/frame given by that anim dispatch structure
};


#define ANIMCMD_NUMMAX_PARAMS                           4

#define ANIMCMD_POSITIONREF3_3                          1
#define ANIMCMD_POSITIONREF2_2                          2
#define ANIMCMD_SLAVEANIM_0                             3
#define ANIMCMD_DEATHANIM_0                             4
#define ANIMCMD_PLAYSOUNDONFRAME_2                      5
#define ANIMCMD_MISCACTIONONFRAME_2                     6
#define     ANIMCMD_MISC_REVERTCAMERA                       0
#define     ANIMCMD_MISC_UNK0                               1
#define     ANIMCMD_MISC_UNK6                               2
#define     ANIMCMD_MISC_MAKEBUBBLE                         3
#define     ANIMCMD_MISC_UNK7                               4
#define     ANIMCMD_MISC_UNK13                              5
#define     ANIMCMD_MISC_UNK14                              6
#define     ANIMCMD_MISC_UNK12                              7
#define     ANIMCMD_MISC_UNK11                              8
#define     ANIMCMD_MISC_REACTIVATEACTIONKEY                12
#define     ANIMCMD_MISC_UNK8                               13
#define     ANIMCMD_MISC_CAMERACTRL1                        14
#define     ANIMCMD_MISC_CAMERACTRL2                        15
#define     ANIMCMD_MISC_UNK9                               16
#define     ANIMCMD_MISC_UNK10                              17
#define     ANIMCMD_MISC_UNK1                               18
#define     ANIMCMD_MISC_UNK2                               19
#define     ANIMCMD_MISC_LARACHANGECLOTH                    20
#define     ANIMCMD_MISC_UNK3                               21
#define     ANIMCMD_MISC_UNK4                               22
#define     ANIMCMD_MISC_HIDEOBJECT                         23
#define     ANIMCMD_MISC_SHOWOBJECT                         24
#define     ANIMCMD_MISC_CAMERACTRL3                        26
#define     ANIMCMD_MISC_ADDITIONALSOUND1                   -32736
#define     ANIMCMD_MISC_UNK5                               32
#define     ANIMCMD_MISC_UNK15                              43
#define     ANIMCMD_MISC_UNK16                              44
#define     ANIMCMD_MISC_ADDITIONALSOUND2                   16416


//
// This class handles the anim commands, which are various commands
// associated with a TR track
//
class CTRAnimCommand
{
public:

    // Constructors / destructors
    CTRAnimCommand(TR2_ANIM_COMMAND* lpsData, DWORD dwFrameStart = -1);
    CTRAnimCommand(TR2_ANIM_COMMAND sCommand, WORD wNumParams) { m_sCommand = sCommand; m_wNumParams = wNumParams; }

    // Accessors
    WORD                wGetNumParams()         const   { return m_wNumParams; }
    TR2_ANIM_COMMAND    sGetCommand()           const   { return m_sCommand; }
    TR2_ANIM_COMMAND    sGetParam(WORD wNum)    const   { ASSERT(wNum < m_wNumParams); return m_sParams[wNum]; }

    void                vSetCommand(TR2_ANIM_COMMAND sCommand);
    void                vSetParam(WORD wNum, TR2_ANIM_COMMAND sValue)
                                                        { ASSERT(wNum < m_wNumParams); m_sParams[wNum] = sValue; }
    // Helpers
    static const bool   boFirstParamIsFrame( const TR2_ANIM_COMMAND sCmd );
    static const DWORD  dwGetNumParams( const TR2_ANIM_COMMAND sCmd );
    const bool          boIsSound() const;

private:

    // Data members
    TR2_ANIM_COMMAND    m_sCommand;                         // command
    WORD                m_wNumParams;                       // num parameters of the command
    TR2_ANIM_COMMAND    m_sParams[ANIMCMD_NUMMAX_PARAMS];   // parameters of the command
};


//
// This class handles the track peculiarities of TR
//
class CTRAnimTrack : public CAnimTrack
{
public:

    // Constructors / destructors
    CTRAnimTrack(DWORD dwNumKeys, DWORD dwNumFrames, DWORD dwFrameRate, FLOAT fAnimFPS);

    virtual ~CTRAnimTrack() { vRelease(); }

    // Accessors
    DWORD   dwGetStateId()                      const   { return m_dwTRStateId; }
    void    vGetPlayParams(short& sTRUnknown, long& lTRSpeed, float& fTRAcceleration) const
                                                        { sTRUnknown = m_sTRUnknown; lTRSpeed = m_lTRSpeed; fTRAcceleration = m_fTRAcceleration; }
	void	vGetAddParams(bitu16& wRealNumAnimCommands, bitu8* lpUnknown2)
														{ wRealNumAnimCommands = m_wRealNumAnimCommands; memcpy(lpUnknown2, m_abUnknown2, 8); }
    DWORD   dwGetNumStateChanges()              const   { return (*m_lpListStateChanges).size(); }
    const CTRStateChange& GetStateChange(DWORD dwNumStateChange) const
                                                        { ASSERT(dwNumStateChange < dwGetNumStateChanges()); return (*m_lpListStateChanges)[dwNumStateChange]; }
    const CTRStateChange& GetStateChangeFromStateId(DWORD dwStateId, DWORD* lpdwIndex = NULL) const;

    DWORD   dwGetNumAnimDispatches()            const;
    DWORD   dwGetNumParamsAnimCommands()        const;

    DWORD   dwGetNumAnimCommands()              const   { return (*m_lpListAnimCommands).size(); }
    const CTRAnimCommand& GetAnimCommand(DWORD dwNumAnimCommand) const
                                                        { ASSERT(dwNumAnimCommand < dwGetNumAnimCommands()); return (*m_lpListAnimCommands)[dwNumAnimCommand]; }
    void            vSetStateId(DWORD dwTRStateId)      { m_dwTRStateId = dwTRStateId; }
    void            vSetPlayParams(short sTRUnknown, long lTRSpeed, float fTRAcceleration)
                                                        { m_sTRUnknown = sTRUnknown; m_lTRSpeed = lTRSpeed; m_fTRAcceleration = fTRAcceleration; }
	void			vSetAddParams(bitu16 wRealNumAnimCommands, bitu8* lpUnknown2)
														{ m_wRealNumAnimCommands = wRealNumAnimCommands; memcpy(&m_abUnknown2, lpUnknown2, 8); }
    void            vPushStateChange(CTRStateChange& StateChange);
    void            vPushAnimCommand(CTRAnimCommand& AnimCommand)
                                                        { (*m_lpListAnimCommands).push_back(AnimCommand); }

    // Operations
    const void      vInitialize();
    const void      vRelease();
    const void      vStealData( CTRAnimTrack& Track );
    const void      vRemoveAnimCommand( const DWORD dwNumAnimCommand);
    const void      vSwapAnimCmde( const DWORD dwAnimCmd1, const DWORD dwAnimCmd2 );
    const void      vRemoveAnimDispatch( const DWORD dwStateId, const DWORD dwTag );

private:

    typedef std::vector<CTRStateChange> TRAnimListStateChanges;
    typedef std::vector<CTRAnimCommand> TRAnimListAnimCommands;

    // Data members
    DWORD                   m_dwTRStateId;              // TR State Id of this track
    short                   m_sTRUnknown;               // ?
    long                    m_lTRSpeed;                 // Translation speed of the root mesh during the anim playing
    float                   m_fTRAcceleration;          // Acceleration of the root mesh during the anim playing
	bitu8					m_abUnknown2[8];			// ?
	bitu16					m_wRealNumAnimCommands;		// real num anim commands as seen in the TR2_ANIMATION structure

    TRAnimListStateChanges* m_lpListStateChanges;       // List of the state changes
    TRAnimListAnimCommands* m_lpListAnimCommands;       // List of the anim commands
};

