#include "stdafx.h"


//----------------------------------------------------------------------------
//
//  Function:   bFindColor
//
//  Synopsis:   Find the nearest color in the TR palette of a given color
//
//  Arguments:  [fR,fG,fB] -- color to look for
//
//  Return:     Index in the palette of the nearest color
//
//----------------------------------------------------------------------------
BYTE    CTRFile::bFindColor(float fR,float fG,float fB)
{
    fR *= 255.0f;
    fG *= 255.0f;
    fB *= 255.0f;

    float       fMin = 1e20f;
    BYTE        bIdx = 0;
    enumTRType  trType = GetTRType();
    TR_FILE&    trFile = GetTRFile();

    for(WORD wI = 0; wI < 256; wI++)
        {
        float fDiff;

        if(trType == TR1_FILE)
            {
            TR2_COLOR col = trFile.tr123File.Palette8.aRGB[wI];

            fDiff = (float)(fabs(fR - col.bRed * 4) + fabs(fG - col.bGreen * 4) + fabs(fB - col.bBlue * 4));
            }
        else
            {
            TR2_COLOR4 col = trFile.tr123File.Palette16.aRGBA[wI];

            fDiff = (float)(fabs(fR - col.bRed) + fabs(fG - col.bGreen) + fabs(fB - col.bBlue));
            }

        if(fDiff < fMin)
            {
            fMin = fDiff;
            bIdx = (BYTE)wI;
            }
        }

    return bIdx;
}


//----------------------------------------------------------------------------
//
//  Function:   vComputeFaceArrayNormals
//
//  Synopsis:   Compute normals at points for a given set of faces
//
//  Arguments:  [lpLstVertices] -- array of points' coordinates
//              [lpLstPtNormals] -- array of points' normals. At the end, this
//               array holds for each point the sum of all face's normal which
//               contain this point
//              [dwNumFaces] -- num faces in the lpArray array
//              [lpArray] -- array of faces
//              [boQuad] -- true if lpArray holds rectangles, false if it
//               holds triangles
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vComputeFaceArrayNormals(TR2_VERTEX* lpLstVertices,vec3* lpLstPtNormals,DWORD dwNumFaces,void* lpArray,bool boQuad)
{
    TR2_FACE3*  lpTriangles     = (TR2_FACE3*) lpArray;
    TR2_FACE4*  lpRectangles    = (TR2_FACE4*) lpArray;

    while(dwNumFaces--)
        {
        bitu16* lpFace  = boQuad ? lpRectangles[dwNumFaces].awVertices : lpTriangles[dwNumFaces].awVertices;

        vec3    Pt1(lpLstVertices[lpFace[0]].sX,lpLstVertices[lpFace[0]].sY,lpLstVertices[lpFace[0]].sZ);
        vec3    Pt2(lpLstVertices[lpFace[1]].sX,lpLstVertices[lpFace[1]].sY,lpLstVertices[lpFace[1]].sZ);
        vec3    Pt3(lpLstVertices[lpFace[2]].sX,lpLstVertices[lpFace[2]].sY,lpLstVertices[lpFace[2]].sZ);
        vec3    Vec1 = Pt2 - Pt1;
        vec3    Vec2 = Pt3 - Pt1;
        vec3    Normal = - (Vec1 ^ Vec2);
        float   fLength = Normal.length();
        
        if(fLength != 0.0f) Normal /= fLength;

        lpLstPtNormals[lpFace[0]] += Normal;
        lpLstPtNormals[lpFace[1]] += Normal;
        lpLstPtNormals[lpFace[2]] += Normal;
        }
}


//----------------------------------------------------------------------------
//
//  Function:   dwGetVersionNumber
//
//  Synopsis:   Get the number version corresponding to a TR sequel
//
//  Arguments:  [TRType] -- TR's type to get the number version
//
//  Return:     Number version
//
//----------------------------------------------------------------------------
DWORD   CTRFile::dwGetVersionNumber(enumTRType TRType)
{
    switch(TRType)
        {
        case TR1_FILE:  return 0x00000020; break;
        case TR2_FILE:  return 0x0000002d; break;
        case TR3_FILE:  return 0xFF180038; break;
        case TR4_FILE:  return 0x00345254; break;
        }

    return 0;
}


//----------------------------------------------------------------------------
//
//  Function:   vComputeObjTextureExtents
//
//  Synopsis:   Compute the (UMin,VMin) (UMax,VMax) of an OBJECT_TEXTURE
//
//  Arguments:  [lpObjText] -- OBJECT_TEXTURE to compute the extents for
//              [bNumVertexPerPoly] -- num vertices of the OBJECT_TEXTURE
//              [boOffset] -- if true, offset the coordinates by the
//               X/YCoordinate field found in the TR2_OBJECT_TEXTURE structure
//               for the computing
//
//  Return:     [*lpsMin / *lpsMinV / *lpsMaxU / *lpsMaxV] -- extents
//
//----------------------------------------------------------------------------
/*static*/ void    CTRFile::vComputeObjTextureExtents(TR2_OBJECT_TEXTURE* lpObjText,BYTE bNumVertexPerPoly,bool boOffset,
            bit16* lpsMinU, bit16* lpsMinV,bit16* lpsMaxU, bit16* lpsMaxV)
{
    bit16 sMinU = 0x7FFF,   sMinV = 0x7FFF;
    bit16 sMaxU = 0x0,      sMaxV = 0x0;

    bit16 sMinU_ = 0x7FFF,   sMinV_ = 0x7FFF;
    bit16 sMaxU_ = 0x0,      sMaxV_ = 0x0;

    for(BYTE bJ = 0; bJ < bNumVertexPerPoly; bJ++)
        {
        bit16 sU = lpObjText->Vertices[bJ].bXPixel;
        bit16 sV = lpObjText->Vertices[bJ].bYPixel;

/*        if(boOffset)
            {
            if( lpObjText->Vertices[bJ].cXCoordinate >= 0) sU++; else sU--;
            if( lpObjText->Vertices[bJ].cYCoordinate >= 0) sV++; else sV--;
            }*/

        if(sMinU > sU) sMinU = sU; if(sMinV > sV) sMinV = sV;
        if(sMaxU < sU) sMaxU = sU; if(sMaxV < sV) sMaxV = sV;

		sU += lpObjText->Vertices[bJ].cXCoordinate;
		sV += lpObjText->Vertices[bJ].cYCoordinate;

        if(sMinU_ > sU) sMinU_ = sU; if(sMinV_ > sV) sMinV_ = sV;
        if(sMaxU_ < sU) sMaxU_ = sU; if(sMaxV_ < sV) sMaxV_ = sV;
        }

	if( sMinU_ <= sMinU )
		{
		if(sMinU_ > 0) sMinU = sMinU_ - 1; else sMinU = 0;

		for(BYTE bJ = 0; bJ < bNumVertexPerPoly; bJ++)
			{
			bit16 sU = lpObjText->Vertices[bJ].bXPixel;

			sU += lpObjText->Vertices[bJ].cXCoordinate;

			if( sU == sMinU_ ) { /*lpObjText->Vertices[bJ].bXPixel = sMinU;*/ lpObjText->Vertices[bJ].cXCoordinate = +1; }
			}
		}
	if( sMinV_ <= sMinV )
		{
		if(sMinV_ > 0) sMinV = sMinV_ - 1; else sMinV = 0;

		for(BYTE bJ = 0; bJ < bNumVertexPerPoly; bJ++)
			{
			bit16 sV = lpObjText->Vertices[bJ].bYPixel;

			sV += lpObjText->Vertices[bJ].cYCoordinate;

			if( sV == sMinV_ ) { /*lpObjText->Vertices[bJ].bYPixel = sMinV;*/ lpObjText->Vertices[bJ].cYCoordinate = +1; }
			}
		}
	if( sMaxU_ >= sMaxU )
		{
		if(sMaxU_ < 255) sMaxU = sMaxU_ + 1; else sMaxU = 255;

		for(BYTE bJ = 0; bJ < bNumVertexPerPoly; bJ++)
			{
			bit16 sU = lpObjText->Vertices[bJ].bXPixel;

			sU += lpObjText->Vertices[bJ].cXCoordinate;

			if( sU == sMaxU_ ) { /*lpObjText->Vertices[bJ].bXPixel = sMaxU;*/ lpObjText->Vertices[bJ].cXCoordinate = -1; }
			}
		}
	if( sMaxV_ >= sMaxV )
		{
		if(sMaxV_ < 255) sMaxV = sMaxV_ + 1; else sMaxV = 255;

		for(BYTE bJ = 0; bJ < bNumVertexPerPoly; bJ++)
			{
			bit16 sV = lpObjText->Vertices[bJ].bYPixel;

			sV += lpObjText->Vertices[bJ].cYCoordinate;

			if( sV == sMaxV_ ) { /*lpObjText->Vertices[bJ].bYPixel = sMaxV;*/ lpObjText->Vertices[bJ].cYCoordinate = -1; }
			}
		}

    if(lpsMinU) *lpsMinU = sMinU;
    if(lpsMaxU) *lpsMaxU = sMaxU;
    if(lpsMinV) *lpsMinV = sMinV;
    if(lpsMaxV) *lpsMaxV = sMaxV;
}


//----------------------------------------------------------------------------
//
//  Function:   dwReplaceColoredPolys
//
//  Synopsis:   Replace all the colored polys in the level by textured ones.
//
//  Arguments:  -
//
//  Return:     number of polys replaced
//
//----------------------------------------------------------------------------
DWORD   CTRFile::dwReplaceColoredPolys(TR2_PALETTE8& Palette8, TR2_PALETTE16& Palette16)
{
    DWORD   dwNumPolys = 0;

    //
    // Build a new textile from the color palette
    //
    DWORD           dwNumTiles  = m_trFile.tr123File.dwNumTextiles;
    TR2_TEXTILE16** lplp16      = new TR2_TEXTILE16* [dwNumTiles + 1];

    memcpy(lplp16, m_trFile.tr123File.lplpTextiles16,  sizeof(TR2_TEXTILE16*) * dwNumTiles);
    lplp16[dwNumTiles] = new TR2_TEXTILE16;
    memset(lplp16[dwNumTiles], 255, sizeof(TR2_TEXTILE16));
    
    if(m_TRType != TR1_FILE)
        {
        DWORD dwLgn     = 0;
        DWORD dwOffset  = 0;

        for(DWORD dwI = 0; dwI < 256; dwI++)
            {
            TR2_COLOR4& color = Palette16.aRGBA[dwI];
            bitu16  wColor = ((color.bRed >> 3) << 10) + ((color.bGreen >> 3) << 5) + (color.bBlue >> 3) + 0x8000;
            for(BYTE bJ = 0; bJ < 3; bJ++)
                for(BYTE bI = 0; bI < 3; bI++)
                    lplp16[dwNumTiles]->awTile[dwLgn * 256 + dwOffset + bI + bJ * 256] = wColor;
            dwOffset += 3;
            if(dwOffset + 3 >= 256)
                {
                dwOffset = 0;
                dwLgn += 3;
                }
            }
        }
    else
        {
        DWORD dwLgn     = 0;
        DWORD dwOffset  = 0;

        for(DWORD dwI = 0; dwI < 256; dwI++)
            {
            TR2_COLOR&  color = Palette8.aRGB[dwI];
            bitu16      wColor = ((color.bRed >> 1) << 10) + ((color.bGreen >> 1) << 5) + (color.bBlue >> 1) + 0x8000;
            for(BYTE bJ = 0; bJ < 3; bJ++)
                for(BYTE bI = 0; bI < 3; bI++)
                    lplp16[dwNumTiles]->awTile[dwLgn * 256 + dwOffset + bI + bJ * 256] = wColor;
            dwOffset += 3;
            if(dwOffset + 3 >= 256)
                {
                dwOffset = 0;
                dwLgn += 3;
                }
            }
        }

    m_trFile.tr123File.dwNumTextiles = dwNumTiles + 1;

    delete [] m_trFile.tr123File.lplpTextiles16;
    m_trFile.tr123File.lplpTextiles16 = lplp16;

    //
    // Build new OBJECT_TEXTURE structures for the 256 colors of the palette
    // There are 2 new sets of 256 OBJECT_TEXTURE structs: one for tris and one for quads
    //
    DWORD   dwNewNumObjText = 256 * 2;
    DWORD   dwNumObjText    = m_trFile.tr123File.dwNumObjectTextures;
    TR2_OBJECT_TEXTURE* lpListOText = new TR2_OBJECT_TEXTURE [ dwNumObjText + dwNewNumObjText ];

    memcpy(lpListOText, m_trFile.tr123File.lpObjectTextures, sizeof(TR2_OBJECT_TEXTURE) * dwNumObjText);

    for(BYTE bJ = 0; bJ < 2; bJ++)
        {
        DWORD dwLgn     = 0;
        DWORD dwOffset  = 0;

        for(DWORD dwI = 0; dwI < 256; dwI++)
            {
            lpListOText[dwNumObjText].wAttribute  = 0;
            lpListOText[dwNumObjText].wTile       = (bitu16)dwNumTiles;
            lpListOText[dwNumObjText].Vertices[0].bXPixel = (bitu8)dwOffset;
            lpListOText[dwNumObjText].Vertices[0].bYPixel = (bitu8)dwLgn;
            lpListOText[dwNumObjText].Vertices[0].cXCoordinate = +1;
            lpListOText[dwNumObjText].Vertices[0].cYCoordinate = +1;
            lpListOText[dwNumObjText].Vertices[1].bXPixel = (bitu8)(dwOffset + 2);
            lpListOText[dwNumObjText].Vertices[1].bYPixel = (bitu8)dwLgn;
            lpListOText[dwNumObjText].Vertices[1].cXCoordinate = -1;
            lpListOText[dwNumObjText].Vertices[1].cYCoordinate = +1;
            lpListOText[dwNumObjText].Vertices[2].bXPixel = (bitu8)(dwOffset + 2);
            lpListOText[dwNumObjText].Vertices[2].bYPixel = (bitu8)(dwLgn + 2);
            lpListOText[dwNumObjText].Vertices[2].cXCoordinate = -1;
            lpListOText[dwNumObjText].Vertices[2].cYCoordinate = -1;
            if(bJ)
                {
                lpListOText[dwNumObjText].Vertices[3].bXPixel = (bitu8)dwOffset;
                lpListOText[dwNumObjText].Vertices[3].bYPixel = (bitu8)(dwLgn + 2);
                lpListOText[dwNumObjText].Vertices[3].cXCoordinate = +1;
                lpListOText[dwNumObjText].Vertices[3].cYCoordinate = -1;
                }
            else
                {
                lpListOText[dwNumObjText].Vertices[3].bXPixel = 0;
                lpListOText[dwNumObjText].Vertices[3].bYPixel = 0;
                lpListOText[dwNumObjText].Vertices[3].cXCoordinate = 0;
                lpListOText[dwNumObjText].Vertices[3].cYCoordinate = 0;
                }
            dwNumObjText++;
            dwOffset += 3;
            if(dwOffset + 3 >= 256)
                {
                dwOffset = 0;
                dwLgn += 3;
                }
            }
        }

    delete [] m_trFile.tr123File.lpObjectTextures;
    m_trFile.tr123File.lpObjectTextures = lpListOText;
    m_trFile.tr123File.dwNumObjectTextures = dwNumObjText;

    //
    // Process the meshes and replace colored polys by textured ones
    //
    for(DWORD dwNumMesh = 0; dwNumMesh < m_trFile.tr123File.dwNumMeshes; dwNumMesh++)
        {
        TR123_MESH* lpMesh = m_trFile.tr123File.lplpMeshes[dwNumMesh];

        dwNumPolys += lpMesh->sNumColoredRectangles + lpMesh->sNumColoredTriangles;

        // Process colored rectangles
        if(lpMesh->sNumColoredRectangles)
            {
            DWORD       dwNewNumRectangles = lpMesh->sNumTexturedRectangles + lpMesh->sNumColoredRectangles;
            TR2_FACE4*  lpList = new TR2_FACE4 [ dwNewNumRectangles ];

            memcpy(lpList, lpMesh->lpTexturedRectangles, sizeof(TR2_FACE4) * lpMesh->sNumTexturedRectangles);

            for(DWORD dwI = 0; dwI < (DWORD)lpMesh->sNumColoredRectangles; dwI++)
                {
                DWORD dwIndex;

                if(m_TRType != TR1_FILE)
                    {
                    dwIndex = lpMesh->lpColoredRectangles[dwI].wTextureIndex >> 8;
                    }
                else
                    {
                    dwIndex = lpMesh->lpColoredRectangles[dwI].wTextureIndex & 0xFF;
                    }

                lpList[lpMesh->sNumTexturedRectangles + dwI] = lpMesh->lpColoredRectangles[dwI];
                lpList[lpMesh->sNumTexturedRectangles + dwI].wTextureIndex = (bitu16)(dwNumObjText - 256 + dwIndex);
                }

            delete [] lpMesh->lpTexturedRectangles;
            delete [] lpMesh->lpColoredRectangles;

            lpMesh->lpTexturedRectangles    = lpList;
            lpMesh->lpColoredRectangles     = NULL;
            lpMesh->sNumTexturedRectangles  = (bit16)dwNewNumRectangles;
            lpMesh->sNumColoredRectangles   = 0;
            }

        // Process colored triangles
        if(lpMesh->sNumColoredTriangles)
            {
            DWORD       dwNewNumTriangles = lpMesh->sNumTexturedTriangles + lpMesh->sNumColoredTriangles;
            TR2_FACE3*  lpList = new TR2_FACE3 [ dwNewNumTriangles ];

            memcpy(lpList, lpMesh->lpTexturedTriangles, sizeof(TR2_FACE3) * lpMesh->sNumTexturedTriangles);

            for(DWORD dwI = 0; dwI < (DWORD)lpMesh->sNumColoredTriangles; dwI++)
                {
                DWORD dwIndex;

                if(m_TRType != TR1_FILE)
                    {
                    dwIndex = lpMesh->lpColoredTriangles[dwI].wTextureIndex >> 8;
                    }
                else
                    {
                    dwIndex = lpMesh->lpColoredTriangles[dwI].wTextureIndex & 0xFF;
                    }

                lpList[lpMesh->sNumTexturedTriangles + dwI] = lpMesh->lpColoredTriangles[dwI];
                lpList[lpMesh->sNumTexturedTriangles + dwI].wTextureIndex = (bitu16)(dwNumObjText - 512 + dwIndex);
                }

            delete [] lpMesh->lpTexturedTriangles;
            delete [] lpMesh->lpColoredTriangles;

            lpMesh->lpTexturedTriangles     = lpList;
            lpMesh->lpColoredTriangles      = NULL;
            lpMesh->sNumTexturedTriangles   = (bit16)dwNewNumTriangles;
            lpMesh->sNumColoredTriangles    = 0;
            }
        }

    return dwNumPolys;
}


/*static*/ MMRESULT CTRFile::dwConvertADPCMToPCM( WAVEFORMATEX& wfxSrc, WAVEFORMATEX& wfxDst, DWORD dwCompressedSize, BYTE* lpbSrc, DWORD& dwUncompressedSize, BYTE* lpbDst )
{
	HACMSTREAM		has;

	MMRESULT	res = acmStreamOpen( &has, NULL, &wfxSrc, &wfxDst, NULL, NULL, NULL, 0 );

	if( res != 0 )
		{
		return res;
		}

	ACMSTREAMHEADER	ash;

	ZeroMemory( &ash, sizeof( ash ) );

	ash.cbStruct = sizeof( ash );
	ash.pbSrc = lpbSrc;
	ash.cbSrcLength = dwCompressedSize;
	ash.pbDst = lpbDst;
	ash.cbDstLength = dwUncompressedSize;

	res = acmStreamPrepareHeader( has, &ash, 0 );

	if( res != 0 )
		{
		acmStreamClose( has, 0 );

		return res;
		}

	res = acmStreamConvert( has, &ash, 0 );

	if( res != 0 )
		{
		acmStreamUnprepareHeader( has, &ash, 0 );
		acmStreamClose( has, 0 );

		return res;
		}

	_ASSERTE( ash.cbSrcLength == ash.cbSrcLengthUsed );

	dwUncompressedSize = ash.cbDstLengthUsed;

	acmStreamUnprepareHeader( has, &ash, 0 );
	acmStreamClose( has, 0 );

	return 0;
}


/*static */ BYTE*	CTRFile::lpbUncompressSound( TR4_SOUND* lpSound, DWORD& dwUncompressedSize )
{
	BYTE* lpb = lpSound->lpbData;

	WAVEFORMATEX*	lpwfxSrc =(WAVEFORMATEX*) ( lpb + 20 );
	WAVEFORMATEX	wfxDst = *lpwfxSrc;

	wfxDst.wFormatTag = 1;	// not compressed;
	wfxDst.nSamplesPerSec = 22050;
	wfxDst.wBitsPerSample = 16;
	wfxDst.nBlockAlign = wfxDst.nChannels * wfxDst.wBitsPerSample / 8;
	wfxDst.nAvgBytesPerSec = wfxDst.nSamplesPerSec * wfxDst.nBlockAlign;
	wfxDst.cbSize = 0;

	BYTE bHeader[] = {
				0x52, 0x49, 0x46, 0x46, 0, 0, 0, 0, 0x57, 0x41, 0x56, 0x45, 0x66, 0x6d, 0x74, 0x20, 0x10, 0, 0, 0,	// 20
				0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, // 16
				0x64, 0x61, 0x74, 0x61, 0, 0, 0, 0 // 8
				 };

	DWORD	dwUnk = lpSound->dwSoundLen * 4;

	BYTE*	lpbSrc = lpb + 0x5A;
	DWORD	dwSrcSize = lpSound->dwSoundLen - 0x5A;
	BYTE*	lpbDst = new BYTE [ dwUnk ];
	DWORD	dwDstSize = dwUnk - sizeof( bHeader );

	memcpy( lpbDst, bHeader, sizeof( bHeader ) );
	memcpy( lpbDst + 20, &wfxDst, sizeof( wfxDst ) - 2 );

	if( dwConvertADPCMToPCM( *lpwfxSrc, wfxDst, dwSrcSize, lpbSrc, dwDstSize, lpbDst + sizeof( bHeader ) ) == 0 )
		{
		* ( (DWORD*) &lpbDst[ 4 ] ) = dwDstSize + sizeof( bHeader ) - 8;
		* ( (DWORD*) &lpbDst[ 40 ] ) = dwDstSize;

		dwUncompressedSize = dwDstSize + sizeof( bHeader );
		}
	else
		{
		delete [] lpbDst;

		lpbDst = NULL;
		}

	return lpbDst;
}
